import re

from .const import Bound, inf
from .interval import Interval


def from_string(string, conv, *, bound=r'.+?', disj=r' ?\| ?', sep=r', ?',
                left_open=r'\(', left_closed=r'\[', right_open=r'\)', right_closed=r'\]',
                pinf=r'\+inf', ninf=r'-inf'):
    """
    Parse given string and create an Interval instance.
    A converter function has to be provided to convert a bound (as string) to a value.

    :param string: string to parse.
    :param conv: function that converts a bound (as string) to an object.
    :param bound: pattern that matches a value.
    :param disj: pattern that matches the disjunctive operator (default matches '|').
    :param sep: pattern that matches a bounds separator (default matches ',').
    :param left_open: pattern that matches a left open boundary (default matches '(').
    :param left_closed: pattern that matches a left closed boundary (default matches '[').
    :param right_open: pattern that matches a right open boundary (default matches ')').
    :param right_closed: pattern that matches a right closed boundary (default matches ']').
    :param pinf: pattern that matches a positive infinity (default matches '+inf').
    :param ninf: pattern that matches a negative infinity (default matches '-inf').
    :return: an Interval instance.
    """

    re_left_boundary = r'(?P<left>{}|{})'.format(left_open, left_closed)
    re_right_boundary = r'(?P<right>{}|{})'.format(right_open, right_closed)
    re_bounds = r'(?P<lower>{bound})({sep}(?P<upper>{bound}))?'.format(bound=bound, sep=sep)
    re_interval = r'{}(|{}){}'.format(re_left_boundary, re_bounds, re_right_boundary)
    re_intervals = r'{}(?P<disj>{})?'.format(re_interval, disj)

    intervals = []
    has_more = True

    def _convert(bound):
        if re.match(pinf, bound):
            return inf
        elif re.match(ninf, bound):
            return -inf
        else:
            return conv(bound)

    while has_more:
        match = re.match(re_intervals, string)
        if match is None:
            has_more = False
        else:
            group = match.groupdict()

            left = Bound.CLOSED if re.match(left_closed + '$', group['left']) else Bound.OPEN
            right = Bound.CLOSED if re.match(right_closed + '$', group['right']) else Bound.OPEN

            lower = group.get('lower', None)
            upper = group.get('upper', None)
            lower = _convert(lower) if lower is not None else inf
            upper = _convert(upper) if upper is not None else lower

            intervals.append(Interval.from_atomic(left, lower, upper, right))
            string = string[match.end():]

    return Interval(*intervals)


def to_string(interval, conv=repr, *, disj=' | ', sep=',', left_open='(',
              left_closed='[', right_open=')', right_closed=']', pinf='+inf', ninf='-inf'):
    """
    Export given interval to string.

    :param interval: an interval.
    :param conv: function that is used to represent a bound (default is `repr`).
    :param disj: string representing disjunctive operator (default is ' | ').
    :param sep: string representing bound separator (default is ',').
    :param left_open: string representing left open boundary (default is '(').
    :param left_closed: string representing left closed boundary (default is '[').
    :param right_open: string representing right open boundary (default is ')').
    :param right_closed: string representing right closed boundary (default is ']').
    :param pinf: string representing a positive infinity (default is '+inf').
    :param ninf: string representing a negative infinity (default is '-inf').
    :return: a string representation for given interval.
    """
    if interval.empty:
        return '{}{}'.format(left_open, right_open)

    def _convert(bound):
        if bound == inf:
            return pinf
        elif bound == -inf:
            return ninf
        else:
            return conv(bound)

    exported_intervals = []
    for item in interval:
        left = left_open if item.left == Bound.OPEN else left_closed
        right = right_open if item.right == Bound.OPEN else right_closed

        lower = _convert(item.lower)
        upper = _convert(item.upper)

        if item.lower == item.upper:
            exported_intervals.append('{}{}{}'.format(left, lower, right))
        else:
            exported_intervals.append('{}{}{}{}{}'.format(left, lower, sep, upper, right))

    return disj.join(exported_intervals)


def from_data(data, conv=None, *, pinf=float('inf'), ninf=float('-inf')):
    """
    Import an interval from a piece of data.

    :param data: a list of 4-uples (left, lower, upper, right).
    :param conv: function that converts "lower" and "upper" to bounds, default to identity.
    :param pinf: value used to represent positive infinity.
    :param ninf: value used to represent negative infinity.
    :return: an Interval instance.
    """
    intervals = []
    conv = (lambda v: v) if conv is None else conv

    def _convert(bound):
        if bound == pinf:
            return inf
        elif bound == ninf:
            return -inf
        else:
            return conv(bound)

    for item in data:
        left, lower, upper, right = item
        intervals.append(Interval.from_atomic(
            Bound(left),
            _convert(lower),
            _convert(upper),
            Bound(right),
        ))
    return Interval(*intervals)


def to_data(interval, conv=None, *, pinf=float('inf'), ninf=float('-inf')):
    """
    Export given interval to a list of 4-uples (left, lower,
    upper, right).

    :param interval: an interval.
    :param conv: function that convert bounds to "lower" and "upper", default to identity.
    :param pinf: value used to encode positive infinity.
    :param ninf: value used to encode negative infinity.
    :return: a list of 4-uples (left, lower, upper, right)
    """
    conv = (lambda v: v) if conv is None else conv

    data = []

    def _convert(bound):
        if bound == inf:
            return pinf
        elif bound == -inf:
            return ninf
        else:
            return conv(bound)

    for item in interval:
        data.append((
            item.left.value,
            _convert(item.lower),
            _convert(item.upper),
            item.right.value
        ))
    return data



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

programName: str = "BORIS"

ETHOGRAM_REPOSITORY_URL = "https://www.boris.unito.it/static/ethograms"

project_format_version: str = "7.0"

SECONDS_PER_DAY: int = 86_400

# cutoff value for displaying HH:MM:SS.zzz or YYYY-mm-DD HH:MM:SS:ZZZ
HOUR_CUTOFF: int = 7 * 24
DATE_CUTOFF: int = HOUR_CUTOFF * 60 * 60  # 1 week

SMART_TIME_CUTOFF_DEFAULT: int = 300

# minimal project version for handling observations from images
IMAGES_OBS_PROJECT_MIN_VERSION: tuple = (8, 6)

IMAGE_EXTENSIONS: tuple = ("*.jpg", "*.png", "*.jpeg", "*.tiff", "*.tif", "*.webp", "*.bmp")

CHECK_NEW_VERSION_DELAY: int = 15 * 24 * 60 * 60

N_PLAYER: int = 8

MAX_UNDO_QUEUE: int = 25

NA: str = "NA"

REALTIME_PLOT_CURSOR_COLOR: str = "red"


CTRL_KEY: str = "Ctrl"
ALT_KEY: str = "Alt"

SPECTRO_TIMER: int = 500

function_keys = {
    16777264: "F1",
    16777265: "F2",
    16777266: "F3",
    16777267: "F4",
    16777268: "F5",
    16777269: "F6",
    16777270: "F7",
    16777271: "F8",
    16777272: "F9",
    16777273: "F10",
    16777274: "F11",
    16777275: "F12",
}

PROJECT_NAME = "project_name"
PROJECT_DATE = "project_date"
PROJECT_DESCRIPTION = "project_description"
PROJECT_VERSION = "project_format_version"

TIME_FORMAT = "time_format"

OBSERVATIONS = "observations"
EVENTS = "events"
TIME_OFFSET = "time offset"

CODING_MAP = "coding_map"
BEHAVIORS_CODING_MAP = "behaviors_coding_map"
SUBJECTS = "subjects_conf"
ETHOGRAM = "behaviors_conf"
BEHAVIORAL_CATEGORIES = "behavioral_categories"
BEHAVIORAL_CATEGORIES_CONF = "behavioral_categories_config"
CONVERTERS = "converters"

OFFSET = "offset"

OBSERVATION_TIME_INTERVAL = "observation time interval"

SUBJECT_NAME = "name"

TIME_BUDGET_FORMAT = "time_budget_format"
"""
COMPACT_TIME_BUDGET_FORMAT = "compact"
LONG_TIME_BUDGET_FORMAT = "long"
DEFAULT_TIME_BUDGET_FORMAT = LONG_TIME_BUDGET_FORMAT
"""

DESCRIPTION = "description"

TIME_BIN_SIZE = "time bin size"

CODING_MAP_RESIZE_W = 640
CODING_MAP_RESIZE_H = 640

# playerType
LIVE = "LIVE"
MEDIA = "MEDIA"
IMAGES = "IMAGES"

VIEWER_MEDIA = "VIEWER_MEDIA"
VIEWER_LIVE = "VIEWER_LIVE"
VIEWER_IMAGES = "VIEWER_IMAGES"

VIEWERS = (VIEWER_MEDIA, VIEWER_LIVE, VIEWER_IMAGES)

POINT_EVENT_PLOT_DURATION = 0.010
POINT_EVENT_PLOT_COLOR = "black"

CHAR_FORBIDDEN_IN_MODIFIERS = "(|),`~"

ADAPT_FAST_JUMP = "adapt_fast_jump"
ADAPT_FAST_JUMP_DEFAULT = False


subjectsFields = ["key", SUBJECT_NAME, DESCRIPTION]

UNPAIRED = "UNPAIRED"
DISPLAY_SUBTITLES = "display_subtitles"


YES = "Yes"
NO = "No"
CANCEL = "Cancel"
APPEND = "Append"
REPLACE = "Replace"
REMOVE = "Remove"
SAVE = "Save"
DISCARD = "Discard"
OK = "OK"
OVERWRITE = "Overwrite"
OVERWRITE_ALL = "Overwrite all"
SKIP = "Skip"
SKIP_ALL = "Skip all"

NO_FOCAL_SUBJECT = "No focal subject"

TYPE = "type"
FILE = "file"
COLOR = "color"
DIRECTORIES_LIST = "directories_list"

SAVE_DATASET = 32

PLOT_DATA = "plot_data"

PLOT_DATA_FILEPATH_IDX = 0
PLOT_DATA_COLUMNS_IDX = 1
PLOT_DATA_PLOTTITLE_IDX = 2
PLOT_DATA_VARIABLENAME_IDX = 3
PLOT_DATA_CONVERTERS_IDX = 4
PLOT_DATA_TIMEINTERVAL_IDX = 5
PLOT_DATA_TIMEOFFSET_IDX = 6
PLOT_DATA_SUBSTRACT1STVALUE_IDX = 7
PLOT_DATA_PLOTCOLOR_IDX = 8

DATA_PLOT_FIELDS = {
    PLOT_DATA_FILEPATH_IDX: "file_path",
    PLOT_DATA_COLUMNS_IDX: "columns",
    PLOT_DATA_PLOTTITLE_IDX: "title",
    PLOT_DATA_VARIABLENAME_IDX: "variable_name",
    PLOT_DATA_CONVERTERS_IDX: "converters",
    PLOT_DATA_TIMEINTERVAL_IDX: "time_interval",
    PLOT_DATA_TIMEOFFSET_IDX: "time_offset",
    PLOT_DATA_SUBSTRACT1STVALUE_IDX: "substract_first_value",
    PLOT_DATA_PLOTCOLOR_IDX: "color",
}
DATA_PLOT_STYLES = ["b-", "r-", "g-", "bo", "ro", "go"]

TIME = "time"
BEHAVIOR_CODE = "code"
SUBJECT = "subject"
MODIFIER = "modifier"
COMMENT = "comment"
BEHAVIOR_KEY = "key"
BEHAVIOR_CATEGORY = "category"

FILE_PATH = "file_path"

IMAGE_INDEX = "image index"
IMAGE_PATH = "image path"
FRAME_INDEX = "frame index"

ZOOM_LEVEL = "zoom level"
PAN_X = "pan-x"
PAN_Y = "pan-y"
ROTATION_ANGLE = "rotation angle"
DISPLAY_MEDIA_SUBTITLES = "display subtitles"
OVERLAY = "video overlay"


USE_EXIF_DATE = "use_exif_date"
TIME_LAPSE = "time_lapse_delay"


# fields for event configuration
"""
fields = {
    "type": 0,
    "key": 1,
    "code": 2,
    "description": 3,
    "modifiers": 4,
    "excluded": 5,
    "coding map": 6,
}
"""

ETHOGRAM_TABLE_COLUMNS: dict = {
    0: "key",
    1: "code",
    2: "type",
    3: "description",
    4: "color",
    5: "category",
    6: "modifiers",
    7: "excluded",
}

# fields in ethogram table from project window
behavioursFields: dict = {
    "type": 0,
    "key": 1,
    "code": 2,
    "description": 3,
    "color": 4,
    "category": 5,
    "modifiers": 6,
    "excluded": 7,
    "coding map": 8,
}
"""
ETHOGRAM_FIELDS = [
    "type",
    "key",
    "code",
    "description",
    "category",
    "modifiers",
    "excluded",
    "coding map",
]
"""
ETHOGRAM_EDITABLE_FIELDS: tuple = ("key", "code", "description")

PROJECT_BEHAVIORS_KEY_FIELD_IDX = 1
PROJECT_BEHAVIORS_CODE_FIELD_IDX = 2

MEDIA_FILE_PATH_IDX = 2
HAS_AUDIO_IDX = 6

STATE_EVENT = "State event"
POINT_EVENT = "Point event"

BEHAVIOR_TYPES = [
    "Point event",
    "State event",
    "Point event with coding map",
    "State event with coding map",
]

DEFAULT_BEHAVIOR_TYPE = "Point event"

MEDIA_TW_EVENTS_FIELDS_DEFAULT = ("time", FRAME_INDEX, "subject", "code", "type", "modifier", "comment")


# fields for events tablewidget
MEDIA_TW_EVENTS_FIELDS = ("time", FRAME_INDEX, "subject", "code", "type", "modifier", "comment")
# MEDIA_TW_EVENTS_FIELDS = ("time", "subject", "code", "type", "modifier", "comment")
LIVE_TW_EVENTS_FIELDS = ("time", "subject", "code", "type", "modifier", "comment")
IMAGES_TW_EVENTS_FIELDS = ("time", "subject", "code", "type", "modifier", "comment", IMAGE_INDEX, IMAGE_PATH)

TW_EVENTS_FIELDS = {
    MEDIA: MEDIA_TW_EVENTS_FIELDS,
    LIVE: LIVE_TW_EVENTS_FIELDS,
    VIEWER_MEDIA: MEDIA_TW_EVENTS_FIELDS,
    VIEWER_LIVE: LIVE_TW_EVENTS_FIELDS,
    IMAGES: IMAGES_TW_EVENTS_FIELDS,
    VIEWER_IMAGES: IMAGES_TW_EVENTS_FIELDS,
}

# create dictionary index
TW_OBS_FIELD = {}
for observation_type in TW_EVENTS_FIELDS:
    TW_OBS_FIELD[observation_type] = {}
    for idx, field in enumerate(TW_EVENTS_FIELDS[observation_type]):
        TW_OBS_FIELD[observation_type][field] = idx


# fields for project events list
MEDIA_PJ_EVENTS_FIELDS = (TIME, "subject", "code", "modifier", "comment", FRAME_INDEX)
LIVE_PJ_EVENTS_FIELDS = (TIME, "subject", "code", "modifier", "comment")
IMAGES_PJ_EVENTS_FIELDS = (TIME, "subject", "code", "modifier", "comment", IMAGE_INDEX, IMAGE_PATH)

PJ_EVENTS_FIELDS = {
    MEDIA: MEDIA_PJ_EVENTS_FIELDS,
    VIEWER_MEDIA: MEDIA_PJ_EVENTS_FIELDS,
    LIVE: LIVE_PJ_EVENTS_FIELDS,
    VIEWER_LIVE: LIVE_PJ_EVENTS_FIELDS,
    IMAGES: IMAGES_PJ_EVENTS_FIELDS,
    VIEWER_IMAGES: IMAGES_PJ_EVENTS_FIELDS,
}


PJ_OBS_FIELDS = {}
for observation_type in PJ_EVENTS_FIELDS:
    PJ_OBS_FIELDS[observation_type] = {}
    for idx, field in enumerate(PJ_EVENTS_FIELDS[observation_type]):
        PJ_OBS_FIELDS[observation_type][field] = idx


# fields for independent variable definition
tw_indVarFields = ["label", "description", "type", "default value", "possible values"]


EVENT_TIME_FIELD_IDX = 0
EVENT_SUBJECT_FIELD_IDX = 1
EVENT_BEHAVIOR_FIELD_IDX = 2
EVENT_MODIFIER_FIELD_IDX = 3
EVENT_COMMENT_FIELD_IDX = 4
EVENT_STATUS_FIELD_IDX = -1
# EVENT_IMAGEIDX_FIELD_IDX = 6
# EVENT_IMAGEPATH_FIELD_IDX = 7


BEHAV_CODING_MAP_FIELDS = ["name", "Behavior codes"]

# characters not allowed in Excel sheet name
EXCEL_FORBIDDEN_CHARACTERS: str = r"\/*[]:?"


# indexes of project window
MEDIA_TAB_IDX = 0
LIVE_TAB_IDX = 1


HHMMSS = "hh:mm:ss"
HHMMSSZZZ = "hh:mm:ss.zzz"
S = "s"

START_FROM_CURRENT_TIME = "start_from_current_time"
START_FROM_CURRENT_EPOCH_TIME = "start_from_current_epoch_time"

SCAN_SAMPLING_TIME = "scan_sampling_time"

POINT_OBJECT = "Point"
SEGMENT_OBJECT = "Segment"
ANGLE_OBJECT = "Angle"
ORIENTED_ANGLE_OBJECT = "Oriented angle"
POLYGON_OBJECT = "Polygon"
POLYLINE_OBJECT = "Polyline"


NEW = "new"
LIST = "list"
EDIT = "edit"
OPEN = "open"
VIEW = "view"
OBS_START = "start"
SELECT = "select"
SINGLE = "single"
MULTIPLE = "multiple"


SELECT1 = "select1"

FILTERED_BEHAVIORS = "filtered behaviors"

SELECTED_BEHAVIORS = "selected behaviors"
SELECTED_SUBJECTS = "selected subjects"
INCLUDE_MODIFIERS = "include modifiers"
EXCLUDE_BEHAVIORS = "exclude behaviors"
EXCLUDED_BEHAVIORS = "excluded behaviors"
TIME_INTERVAL = "time"
START_TIME = "start time"
END_TIME = "end time"

# indep variables
NUMERIC = "numeric"
NUMERIC_idx = 0
TEXT = "text"
TEXT_idx = 1
SET_OF_VALUES = "value from set"
SET_OF_VALUES_idx = 2
TIMESTAMP = "timestamp"
TIMESTAMP_idx = 3

TIME_FULL_OBS = "full obs"
TIME_EVENTS = "limit to events"
TIME_ARBITRARY_INTERVAL = "time interval"

AVAILABLE_INDEP_VAR_TYPES = [NUMERIC, TEXT, SET_OF_VALUES, TIMESTAMP]

INDEPENDENT_VARIABLES = "independent_variables"
OBSERVATIONS = "observations"

CLOSE_BEHAVIORS_BETWEEN_VIDEOS = "close_behaviors_between_videos"

# MPV hardware decode
MPV_HWDEC = "mpv_hwdec"
MPV_HWDEC_NO = "no"
MPV_HWDEC_AUTO = "auto"
MPV_HWDEC_AUTOSAFE = "auto-safe"
MPV_HWDEC_OPTIONS = (MPV_HWDEC_AUTO, MPV_HWDEC_AUTOSAFE, MPV_HWDEC_NO)
MPV_HWDEC_DEFAULT_VALUE = MPV_HWDEC_AUTO


PROJECT_FILE_INDENTATION = "project file indentation"
PROJECT_FILE_INDENTATION_COMBO_OPTIONS = ("None", "Newline", "Tab", "2 spaces", "4 spaces")
PROJECT_FILE_INDENTATION_OPTIONS = (None, 0, "\t", 2, 4)
PROJECT_FILE_INDENTATION_DEFAULT_VALUE = None

TOOLBAR_ICON_SIZE = "toolbar icon size"
DEFAULT_TOOLBAR_ICON_SIZE_VALUE = 24

DARK_MODE = "dark_mode"
DARK_MODE_DEFAULT_VALUE = False

VIDEO_VIEWER = 0
PICTURE_VIEWER = 1

SAVE_FRAMES = "save_frames"
MEMORY_FOR_FRAMES = "memory_for_frames"
DEFAULT_MEMORY_FOR_FRAMES = 80  # % total memory
DISK = "disk"
MEMORY = "memory"
DEFAULT_FRAME_MODE = DISK

MEDIA_FILE_INFO = "media_file_info"
MEDIA_INFO = "media_info"
LENGTH = "length"
FPS = "fps"
HAS_AUDIO = "hasAudio"
HAS_VIDEO = "hasVideo"
MEDIA_CREATION_TIME = "media_creation_time"

STATE = "STATE"
POINT = "POINT"

START = "START"
STOP = "STOP"

PLAYER1, PLAYER2 = "1", "2"
ALL_PLAYERS = [str(x + 1) for x in range(N_PLAYER)]

VISUALIZE_SPECTROGRAM = "visualize_spectrogram"
VISUALIZE_WAVEFORM = "visualize_waveform"
MEDIA_CREATION_DATE_AS_OFFSET = "media_creation_date_as_offset"

MEDIA_SCAN_SAMPLING_DURATION = "media_scan_sampling_duration"
IMAGE_DISPLAY_DURATION = "image_display_duration"

# plot type
WAVEFORM_PLOT = "waveform"
SPECTROGRAM_PLOT = "spectrogram"
EVENTS_PLOT = "plot_events"


POINT_EVENT_ST_DURATION = 0.5

VIDEO_TAB = 0
FRAME_TAB = 1

SLIDER_MAXIMUM = 1000

FRAME_DEFAULT_CACHE_SIZE = 1

# modifiers
MODIFIERS = "modifiers"
SINGLE_SELECTION = 0
MULTI_SELECTION = 1
NUMERIC_MODIFIER = 2
EXTERNAL_DATA_MODIFIER = 3

MODIFIERS_STR = {
    SINGLE_SELECTION: "Single item selection",
    MULTI_SELECTION: "Multiple items selection",
    NUMERIC_MODIFIER: "Numeric",
    EXTERNAL_DATA_MODIFIER: "Value from external data file",
}

# colors
subtitlesColors = [
    "cyan",
    "red",
    "blue",
    "yellow",
    "fuchsia",
    "orange",
    "lime",
    "green",
]

CATEGORY_COLORS_LIST = [
    "#FF96CC",
    "#96FF9C",
    "#CCFFFE",
    "#EEFF70",
    "#FF4F64",
    "#F8BF15",
    "#3DC7AD",
]

CODING_PAD_CONFIG = "coding pad configuration"
CODING_PAD_GEOMETRY = "coding pad geometry"
NO_COLOR_CODING_PAD = "#777777"

SPECTROGRAM_COLOR_MAPS = ["viridis", "inferno", "plasma", "magma", "gray", "YlOrRd"]
SPECTROGRAM_DEFAULT_COLOR_MAP = "viridis"
SPECTROGRAM_DEFAULT_TIME_INTERVAL = 10

# see matplotlib.colors.cnames.keys()
# https://xkcd.com/color/rgb/

# sage colors are no more available
# darksage #598556
# lightsage #bcecac
# sage #87ae73
ACTIVE_MEASUREMENTS_COLOR = "lime"
PASSIVE_MEASUREMENTS_COLOR = "red"

# see matplotlib for color name
BEHAVIORS_PLOT_COLORS = [
    "tab:blue",
    "tab:orange",
    "tab:green",
    "tab:red",
    "tab:purple",
    "tab:brown",
    "tab:pink",
    "tab:gray",
    "tab:olive",
    "tab:cyan",
    "blue",
    "green",
    "red",
    "cyan",
    "magenta",
    "yellow",
    "lime",
    "darksalmon",
    "purple",
    "orange",
    "maroon",
    "silver",
    "slateblue",
    "hotpink",
    "steelblue",
    "darkgoldenrod",
    "aqua",
    "aquamarine",
    "beige",
    "bisque",
    "black",
    "blanchedalmond",
    "blueviolet",
    "brown",
    "burlywood",
    "cadetblue",
    "chartreuse",
    "chocolate",
    "coral",
    "cornflowerblue",
    "cornsilk",
    "crimson",
    "darkblue",
    "darkcyan",
    "darkgreen",
    "darkgrey",
    "darkkhaki",
    "darkmagenta",
    "darkolivegreen",
    "darkorange",
    "darkorchid",
    "darkred",
    "#598556",
    "darkseagreen",
    "darkslateblue",
    "darkslategray",
    "darkslategrey",
    "darkturquoise",
    "darkviolet",
    "deeppink",
    "deepskyblue",
    "dimgray",
    "dimgrey",
    "dodgerblue",
    "firebrick",
    "floralwhite",
    "forestgreen",
    "fuchsia",
    "gainsboro",
    "gold",
    "goldenrod",
    "gray",
    "greenyellow",
    "grey",
    "honeydew",
    "indianred",
    "indigo",
    "khaki",
    "lawngreen",
    "lemonchiffon",
    "lightblue",
    "lightcoral",
    "lightgoldenrodyellow",
    "lightgray",
    "lightgreen",
    "lightgrey",
    "lightpink",
    "#bcecac",
    "lightsalmon",
    "lightseagreen",
    "lightskyblue",
    "lightslategray",
    "lightslategrey",
    "lightsteelblue",
    "lightyellow",
    "limegreen",
    "linen",
    "mediumaquamarine",
    "mediumblue",
    "mediumorchid",
    "mediumpurple",
    "mediumseagreen",
    "mediumslateblue",
    "mediumspringgreen",
    "mediumturquoise",
    "mediumvioletred",
    "midnightblue",
    "mintcream",
    "mistyrose",
    "moccasin",
    "navajowhite",
    "navy",
    "oldlace",
    "olive",
    "olivedrab",
    "orangered",
    "orchid",
    "palegoldenrod",
    "palegreen",
    "paleturquoise",
    "palevioletred",
    "papayawhip",
    "peachpuff",
    "peru",
    "pink",
    "plum",
    "powderblue",
    "rosybrown",
    "royalblue",
    "saddlebrown",
    "#87ae73",
    "salmon",
    "sandybrown",
    "seagreen",
    "seashell",
    "sienna",
    "skyblue",
    "slategray",
    "slategrey",
    "springgreen",
    "tan",
    "teal",
    "thistle",
    "tomato",
    "turquoise",
    "violet",
    "wheat",
    "yellowgreen",
    "darkgray",
]

EMPTY_PROJECT = {
    TIME_FORMAT: HHMMSS,
    PROJECT_DATE: "",
    PROJECT_NAME: "",
    PROJECT_DESCRIPTION: "",
    PROJECT_VERSION: project_format_version,
    SUBJECTS: {},
    ETHOGRAM: {},
    OBSERVATIONS: {},
    BEHAVIORAL_CATEGORIES: [],
    INDEPENDENT_VARIABLES: {},
    CODING_MAP: {},
    BEHAVIORS_CODING_MAP: [],
    CONVERTERS: {},
}

INIT_PARAM = {
    DISPLAY_SUBTITLES: False,
    SAVE_FRAMES: DISK,
    MEMORY_FOR_FRAMES: DEFAULT_MEMORY_FOR_FRAMES,
    ADAPT_FAST_JUMP: ADAPT_FAST_JUMP_DEFAULT,
    # TIME_BUDGET_FORMAT: DEFAULT_TIME_BUDGET_FORMAT,
    MPV_HWDEC: MPV_HWDEC_DEFAULT_VALUE,
    PROJECT_FILE_INDENTATION: PROJECT_FILE_INDENTATION_DEFAULT_VALUE,
    f"{MEDIA} tw fields": MEDIA_TW_EVENTS_FIELDS_DEFAULT,
}

SDIS_EXT = "sds"
TBS_EXT = "tbs"
TSV_EXT = "tsv"
CSV_EXT = "csv"
RDS_EXT = "rds"
PANDAS_DF_EXT = "pkl"
HTML_EXT = "html"
SQL_EXT = "sql"
ODS_EXT = "ods"
XLS_EXT = "xls"
XLSX_EXT = "xlsx"

# Output format
TSV = "Tab Separated Values (*.tsv)"
CSV = "Comma Separated Values (*.csv)"
ODS = "OpenDocument Spreadsheet ODS (*.ods)"
ODS_WB = "OpenDocument Workbook (*.ods)"
XLSX = "Microsoft Excel Spreadsheet XLSX (*.xlsx)"
XLSX_WB = "Microsoft Excel Workbook (*.xlsx)"
XLS = "Legacy Microsoft Excel Spreadsheet XLS (*.xls)"
HTML = "HTML (*.html)"
PANDAS_DF = "Pandas DataFrame (*.pkl)"
RDS = "R dataframe (*.rds)"
SQL = "SQL dump file (*.sql)"
SDIS = "SDIS (*.sds)"
TBS = "Timed Behavioral Sequences (*.tbs)"
TEXT_FILE = "Text file"

FILE_NAME_SUFFIX = {
    TSV: TSV_EXT,
    CSV: CSV_EXT,
    ODS: ODS_EXT,
    ODS_WB: ODS_EXT,
    XLSX: XLSX_EXT,
    XLSX_WB: XLSX_EXT,
    XLS: XLS_EXT,
    HTML: HTML_EXT,
    PANDAS_DF: PANDAS_DF_EXT,
    RDS: RDS_EXT,
    SQL: SQL_EXT,
    SDIS: SDIS_EXT,
    TBS: TBS_EXT,
    TEXT_FILE: "cli",
}



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging
from decimal import Decimal as dec
from typing import Optional

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QCheckBox, QListWidgetItem, QMessageBox

from . import config as cfg
from . import gui_utilities, param_panel, project_functions
from . import utilities as util


def choose_obs_subj_behav_category(
    self,
    selected_observations: list,
    start_coding: Optional[dec] = dec("NaN"),  #  Union[..., None]
    end_coding: Optional[dec] = dec("NaN"),
    maxTime: Optional[dec] = None,
    flagShowIncludeModifiers: bool = True,
    flagShowExcludeBehaviorsWoEvents: bool = True,
    by_category: bool = False,
    n_observations: int = 1,
    show_time_bin_size: bool = False,
    window_title: str = "Select subjects and behaviors",
) -> dict:
    """
    show window for:
    - selection of subjects
    - selection of behaviors (based on selected subjects)
    - selection of time interval
    - inclusion/exclusion of modifiers
    - inclusion/exclusion of behaviors without events (flagShowExcludeBehaviorsWoEvents == True)
    - selection of time bin size (show_time_bin_size == True)

    Args:
        selected_observations (list): List of selected observations
        ...

    Returns:
        dict: {"selected subjects": selectedSubjects,
            "selected behaviors": selectedBehaviors,
            "include modifiers": True/False,
            "exclude behaviors": True/False,
            "time": TIME_FULL_OBS / TIME_EVENTS / TIME_ARBITRARY_INTERVAL
            "start time": startTime,
            "end time": endTime
            }

    """

    paramPanelWindow = param_panel.Param_panel()
    paramPanelWindow.resize(400, 400)
    paramPanelWindow.setWindowTitle(window_title)
    paramPanelWindow.selectedObservations = selected_observations
    paramPanelWindow.pj = self.pj
    paramPanelWindow.extract_observed_behaviors = self.extract_observed_behaviors

    paramPanelWindow.cbIncludeModifiers.setVisible(flagShowIncludeModifiers)
    paramPanelWindow.cbExcludeBehaviors.setVisible(flagShowExcludeBehaviorsWoEvents)
    # show_time_bin_size:
    paramPanelWindow.frm_time_bin_size.setVisible(show_time_bin_size)

    if by_category:
        paramPanelWindow.cbIncludeModifiers.setVisible(False)
        paramPanelWindow.cbExcludeBehaviors.setVisible(False)

    paramPanelWindow.media_duration = maxTime
    paramPanelWindow.start_coding = start_coding
    paramPanelWindow.end_coding = end_coding

    if self.timeFormat == cfg.S:
        paramPanelWindow.start_time.rb_seconds.setChecked(True)
        paramPanelWindow.end_time.rb_seconds.setChecked(True)
    if self.timeFormat == cfg.HHMMSS:
        paramPanelWindow.start_time.rb_time.setChecked(True)
        paramPanelWindow.end_time.rb_time.setChecked(True)

    if n_observations > 1:
        paramPanelWindow.frm_time_interval.setVisible(False)
    else:
        if (start_coding is None) or (start_coding.is_nan()):
            paramPanelWindow.rb_observed_events.setEnabled(False)
            paramPanelWindow.frm_time_interval.setVisible(False)
            paramPanelWindow.rb_user_defined.setVisible(False)
            paramPanelWindow.rb_media_duration.setVisible(False)
        else:
            paramPanelWindow.frm_time_interval.setEnabled(False)
            paramPanelWindow.start_time.set_time(start_coding)
            paramPanelWindow.end_time.set_time(end_coding)

    if selected_observations:
        observedSubjects = project_functions.extract_observed_subjects(self.pj, selected_observations)
    else:
        # load all subjects and "No focal subject"
        observedSubjects = [self.pj[cfg.SUBJECTS][x][cfg.SUBJECT_NAME] for x in self.pj[cfg.SUBJECTS]] + [""]
    selectedSubjects = []

    # add 'No focal subject'
    if "" in observedSubjects:
        selectedSubjects.append(cfg.NO_FOCAL_SUBJECT)
        paramPanelWindow.item = QListWidgetItem(paramPanelWindow.lwSubjects)
        paramPanelWindow.ch = QCheckBox()
        paramPanelWindow.ch.setText(cfg.NO_FOCAL_SUBJECT)
        paramPanelWindow.ch.stateChanged.connect(paramPanelWindow.cb_changed)
        paramPanelWindow.ch.setChecked(True)
        paramPanelWindow.lwSubjects.setItemWidget(paramPanelWindow.item, paramPanelWindow.ch)

    all_subjects = [self.pj[cfg.SUBJECTS][x][cfg.SUBJECT_NAME] for x in util.sorted_keys(self.pj[cfg.SUBJECTS])]

    for subject in all_subjects:
        paramPanelWindow.item = QListWidgetItem(paramPanelWindow.lwSubjects)
        paramPanelWindow.ch = QCheckBox()
        paramPanelWindow.ch.setText(subject)
        paramPanelWindow.ch.stateChanged.connect(paramPanelWindow.cb_changed)
        if subject in observedSubjects:
            selectedSubjects.append(subject)
            paramPanelWindow.ch.setChecked(True)

        paramPanelWindow.lwSubjects.setItemWidget(paramPanelWindow.item, paramPanelWindow.ch)

    logging.debug(f"selected subjects: {selectedSubjects}")

    if selected_observations:
        observedBehaviors = self.extract_observed_behaviors(selected_observations, selectedSubjects)  # not sorted
    else:
        # load all behaviors
        observedBehaviors = [self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in self.pj[cfg.ETHOGRAM]]

    logging.debug(f"observed behaviors: {observedBehaviors}")

    if cfg.BEHAVIORAL_CATEGORIES in self.pj:
        categories = self.pj[cfg.BEHAVIORAL_CATEGORIES][:]
        # check if behavior not included in a category
        try:
            if "" in [
                self.pj[cfg.ETHOGRAM][idx][cfg.BEHAVIOR_CATEGORY]
                for idx in self.pj[cfg.ETHOGRAM]
                if cfg.BEHAVIOR_CATEGORY in self.pj[cfg.ETHOGRAM][idx]
            ]:
                categories += [""]
        except Exception:
            categories = ["###no category###"]

    else:
        categories = ["###no category###"]

    for category in categories:
        if category != "###no category###":
            if category == "":
                paramPanelWindow.item = QListWidgetItem("No category")
                paramPanelWindow.item.setData(34, "No category")
            else:
                paramPanelWindow.item = QListWidgetItem(category)
                paramPanelWindow.item.setData(34, category)

            font = QFont()
            font.setBold(True)
            paramPanelWindow.item.setFont(font)
            paramPanelWindow.item.setData(33, "category")
            paramPanelWindow.item.setData(35, False)

            paramPanelWindow.lwBehaviors.addItem(paramPanelWindow.item)

        for behavior in [self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in util.sorted_keys(self.pj[cfg.ETHOGRAM])]:
            if (categories == ["###no category###"]) or (
                behavior
                in [
                    self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE]
                    for x in self.pj[cfg.ETHOGRAM]
                    if cfg.BEHAVIOR_CATEGORY in self.pj[cfg.ETHOGRAM][x] and self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CATEGORY] == category
                ]
            ):
                paramPanelWindow.item = QListWidgetItem(behavior)
                if behavior in observedBehaviors:
                    paramPanelWindow.item.setCheckState(Qt.Checked)
                else:
                    paramPanelWindow.item.setCheckState(Qt.Unchecked)

                if category != "###no category###":
                    paramPanelWindow.item.setData(33, "behavior")
                    if category == "":
                        paramPanelWindow.item.setData(34, "No category")
                    else:
                        paramPanelWindow.item.setData(34, category)

                paramPanelWindow.lwBehaviors.addItem(paramPanelWindow.item)

    gui_utilities.restore_geometry(paramPanelWindow, "param panel", (600, 500))

    if not paramPanelWindow.exec_():
        return {}

    gui_utilities.save_geometry(paramPanelWindow, "param panel")

    selectedSubjects = paramPanelWindow.selectedSubjects
    selectedBehaviors = paramPanelWindow.selectedBehaviors

    logging.debug(f"selected subjects: {selectedSubjects}")
    logging.debug(f"selected behaviors: {selectedBehaviors}")

    if paramPanelWindow.rb_user_defined.isChecked():
        startTime = paramPanelWindow.start_time.get_time()
        endTime = paramPanelWindow.end_time.get_time()

        if startTime > endTime:
            QMessageBox.warning(
                None,
                cfg.programName,
                "The start time is after the end time",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return {cfg.SELECTED_SUBJECTS: [], cfg.SELECTED_BEHAVIORS: []}

    else:
        startTime = None
        endTime = None

    # if startTime is None:
    #    startTime = dec("NaN")
    # if endTime is None:
    #    endTime = dec("NaN")

    if paramPanelWindow.rb_media_duration.isChecked():
        time_param = cfg.TIME_FULL_OBS
    if paramPanelWindow.rb_observed_events.isChecked():
        time_param = cfg.TIME_EVENTS
    if paramPanelWindow.rb_user_defined.isChecked():
        time_param = cfg.TIME_ARBITRARY_INTERVAL

    return {
        cfg.SELECTED_SUBJECTS: selectedSubjects,
        cfg.SELECTED_BEHAVIORS: selectedBehaviors,
        cfg.INCLUDE_MODIFIERS: paramPanelWindow.cbIncludeModifiers.isChecked(),
        cfg.EXCLUDE_BEHAVIORS: paramPanelWindow.cbExcludeBehaviors.isChecked(),
        "time": time_param,
        cfg.START_TIME: startTime,
        cfg.END_TIME: endTime,
        cfg.TIME_BIN_SIZE: paramPanelWindow.sb_time_bin_size.value(),
    }



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

from PyQt5.QtWidgets import QDialog
from PyQt5.QtCore import pyqtSignal, QEvent
from .video_equalizer_ui import Ui_Equalizer


class Video_equalizer(QDialog, Ui_Equalizer):
    """
    management of video equalizer: brightness, saturation, contrast, gamma and hue
    """

    sendEventSignal = pyqtSignal(int, str, int)
    sendKeyPressSignal = pyqtSignal(QEvent)

    def __init__(self, equalizer, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        self.equalizer = equalizer
        self.pb_close.clicked.connect(self.close)
        self.pb_reset.clicked.connect(self.reset)
        self.pb_reset_all.clicked.connect(self.reset_all)
        self.cb_player.currentIndexChanged.connect(self.player_changed)

        self.hs_brightness.valueChanged.connect(self.parameter_changed)
        self.hs_contrast.valueChanged.connect(self.parameter_changed)
        self.hs_saturation.valueChanged.connect(self.parameter_changed)
        self.hs_gamma.valueChanged.connect(self.parameter_changed)
        self.hs_hue.valueChanged.connect(self.parameter_changed)

        for n_player in self.equalizer:
            self.cb_player.addItem(f"Player {n_player + 1}")

        self.installEventFilter(self)

        self.initialize(0)

    def initialize(self, n_player: int) -> None:
        if n_player not in self.equalizer:
            return

        print(self.equalizer)

        self.hs_brightness.setValue(self.equalizer[n_player]["hs_brightness"])
        self.lb_brightness.setText(str(self.equalizer[n_player]["hs_brightness"]))

        self.hs_contrast.setValue(self.equalizer[n_player]["hs_contrast"])
        self.lb_contrast.setText(str(self.equalizer[n_player]["hs_contrast"]))

        self.hs_saturation.setValue(self.equalizer[n_player]["hs_saturation"])
        self.lb_saturation.setText(str(self.equalizer[n_player]["hs_saturation"]))

        self.hs_gamma.setValue(self.equalizer[n_player]["hs_gamma"])
        self.lb_gamma.setText(str(self.equalizer[n_player]["hs_gamma"]))

        self.hs_hue.setValue(self.equalizer[n_player]["hs_hue"])
        self.lb_hue.setText(str(self.equalizer[n_player]["hs_hue"]))

    def player_changed(self, index):
        self.initialize(index)

    def reset(self):
        """
        Reset value for all parameters to 0 for current player
        """
        for w in (self.hs_brightness, self.hs_contrast, self.hs_saturation, self.hs_gamma, self.hs_hue):
            w.setValue(0)

    def reset_all(self):
        """
        Reset value for all parameters to 0 for all players
        """

        for n_player in self.equalizer:
            for parameter in self.equalizer[n_player]:
                self.equalizer[n_player][parameter] = 0
                self.sendEventSignal.emit(n_player, parameter, 0)

    def parameter_changed(self):
        """
        Send signal when horizontal slider value changed
        """
        if self.cb_player.currentIndex() not in self.equalizer:
            return

        print(type(self.sender().value()))

        self.equalizer[self.cb_player.currentIndex()][self.sender().objectName()] = round(self.sender().value())
        self.sendEventSignal.emit(self.cb_player.currentIndex(), self.sender().objectName(), round(self.sender().value()))

        self.initialize(self.cb_player.currentIndex())

    def eventFilter(self, receiver, event):
        """
        send event (if keypress) to main window
        """
        if event.type() == QEvent.KeyPress:
            self.sendKeyPressSignal.emit(event)
            return True
        else:
            return False


def video_equalizer_show(self):
    """
    Video equalizer
    """

    def update_parameter(n_player, parameter, value):
        """
        update player parameter with value received by signal
        """
        if parameter == "hs_brightness":
            self.dw_player[n_player].player.brightness = value
        if parameter == "hs_contrast":
            self.dw_player[n_player].player.contrast = value
        if parameter == "hs_saturation":
            self.dw_player[n_player].player.saturation = value
        if parameter == "hs_gamma":
            self.dw_player[n_player].player.gamma = value
        if parameter == "hs_hue":
            self.dw_player[n_player].player.hue = value

    # send current parameters
    equalizer = {}
    for n_player, _ in enumerate(self.dw_player):
        equalizer[n_player] = {
            "hs_brightness": round(self.dw_player[n_player].player.brightness),
            "hs_contrast": round(self.dw_player[n_player].player.contrast),
            "hs_saturation": round(self.dw_player[n_player].player.saturation),
            "hs_gamma": round(self.dw_player[n_player].player.gamma),
            "hs_hue": round(self.dw_player[n_player].player.hue),
        }

    self.video_equalizer_wgt = Video_equalizer(equalizer)
    self.video_equalizer_wgt.sendEventSignal.connect(update_parameter)
    # send key press events received by widget to main window
    self.video_equalizer_wgt.sendKeyPressSignal.connect(self.signal_from_widget)

    self.video_equalizer_wgt.show()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""


import os
import tempfile
import pathlib as pl
import logging

from PyQt5.QtWidgets import QFileDialog, QMessageBox, QInputDialog
from PyQt5.QtCore import (
    Qt,
    QProcess,
)

from . import config as cfg
from . import dialog
from . import utilities as util


def ffmpeg_process(self, action: str):
    """
    launch ffmpeg process with QProcess

    Args:
        action (str): "reencode_resize, rotate, merge
    """
    if action not in ("reencode_resize", "rotate", "merge"):
        return

    def readStdOutput(idx):
        """
        read stdout and stderr form qprocess and display them
        """
        self.processes_widget.label.setText(
            (
                "This operation can be long. Be patient...\n"
                "In the meanwhile you can continue to use BORIS\n\n"
                f"Done: {self.processes_widget.number_of_files - len(self.processes)} of {self.processes_widget.number_of_files}"
            )
        )

        # self.processes_widget.lwi.clear()
        std_out = self.processes[idx - 1][0].readAllStandardOutput().data().decode("utf-8")
        if std_out:
            self.processes_widget.lwi.addItems((f"{pl.Path(self.processes[idx - 1][1][2]).name}:   {std_out}",))

        """
        std_err = self.processes[idx - 1][0].readAllStandardError().data().decode("utf-8")
        if std_err:
            self.processes_widget.lwi.addItems((f"{pl.Path(self.processes[idx - 1][1][2]).name}: ERROR: {std_err}",))
            self.flag_ffmpeg_error = True
        """

        self.processes_widget.lwi.scrollToBottom()

    def qprocess_finished(idx):
        """
        function triggered when process finished
        """
        if self.processes:
            del self.processes[idx - 1]
        if self.processes:
            # start new process
            self.processes[-1][0].start(self.processes[-1][1][0], self.processes[-1][1][1])
        else:
            self.processes_widget.label.setText(
                (f"Done: {self.processes_widget.number_of_files - len(self.processes)} of {self.processes_widget.number_of_files}")
            )
            """
            self.processes_widget.hide()
            del self.processes_widget
            """

    if self.processes:
        QMessageBox.warning(self, cfg.programName, "BORIS is already running some job.")
        return

    if action == "merge":
        msg = "Select two or more media files to merge"
        file_type = "Media files (*)"
    else:
        msg = f"Select one or more video files to {action.replace('_', ' and ')}"
        file_type = "Video files (*)"
    fn = QFileDialog().getOpenFileNames(self, msg, "", file_type)
    file_names = fn[0] if type(fn) is tuple else fn

    if not file_names:
        return

    if action == "reencode_resize":
        current_bitrate = 10_000_000  # default 10 Mb/s
        current_resolution = 1024

        r = util.accurate_media_analysis(self.ffmpeg_bin, file_names[0])
        if "error" in r:
            QMessageBox.warning(self, cfg.programName, f"{file_names[0]}. {r['error']}")
        elif r["has_video"]:
            current_bitrate = r.get("bitrate", None)
            if current_bitrate is None:
                current_bitrate = -1
            else:
                current_bitrate = round(current_bitrate / 1024 / 1024)  # Convert to Mb/s
            current_resolution = int(r["resolution"].split("x")[0]) if r["resolution"] is not None else None

        ib = dialog.Input_dialog(
            "Set the parameters for re-encoding / resizing",
            [
                ("sb", "Horizontal resolution (in pixel)", 352, 3840, 100, current_resolution),
                ("sb", "Video quality (bitrate Mb/s)", 1, 1000, 1, current_bitrate),
            ],
        )
        if not ib.exec_():
            return

        if len(file_names) > 1:
            if (
                dialog.MessageDialog(
                    cfg.programName,
                    "All the selected video files will be re-encoded / resized with these parameters",
                    [cfg.OK, cfg.CANCEL],
                )
                == cfg.CANCEL
            ):
                return

        horiz_resol = ib.elements["Horizontal resolution (in pixel)"].value()
        video_quality = ib.elements["Video quality (bitrate Mb/s)"].value()

    if action == "merge":
        if len(file_names) == 1:
            QMessageBox.critical(self, cfg.programName, "Select more than one file")
            return

        file_extensions = []  # check extension of 1st media file
        file_list_lst = []
        for file_name in file_names:
            file_list_lst.append(f"file '{file_name}'")
            file_extensions.append(pl.Path(file_name).suffix)
        if len(set(file_extensions)) > 1:
            QMessageBox.critical(self, cfg.programName, "All media files must have the same format")
            return

        while True:
            output_file_name, _ = QFileDialog().getSaveFileName(self, "Output file name", "", "*")
            if output_file_name == "":
                return
            if pl.Path(output_file_name).suffix != file_extensions[0]:
                QMessageBox.warning(
                    self,
                    cfg.programName,
                    (
                        "The extension of output file must be the same than the extension of input files "
                        f"(<b>{file_extensions[0]}</b>).<br>You selected a {pl.Path(output_file_name).suffix} file."
                    ),
                )
            else:
                break

        # temp file for list of media file to merge
        with tempfile.NamedTemporaryFile() as tmp:
            file_list = tmp.name
        with open(file_list, "w") as f_out:
            f_out.write("\n".join(file_list_lst))

    if action == "rotate":
        rotation_items = ("Rotate 90 clockwise", "Rotate 90 counter clockwise", "rotate 180")

        rotation, ok = QInputDialog.getItem(self, "Rotate media file(s)", "Type of rotation", rotation_items, 0, False)

        if not ok:
            return
        rotation_idx = rotation_items.index(rotation) + 1

    # check if processed files already exist
    if action in ("reencode_resize", "rotate"):
        files_list = []
        for file_name in file_names:
            if action == "reencode_resize":
                fn = f"{file_name}.re-encoded.{horiz_resol}px.{video_quality}Mb.avi"

            if action == "rotate":
                fn = f"{file_name}.rotated{['', '90', '-90', '180'][rotation_idx]}.avi"

            if os.path.isfile(fn):
                files_list.append(fn)

        if files_list:
            response = dialog.MessageDialog(
                cfg.programName,
                "Some file(s) already exist.\n\n" + "\n".join(files_list),
                [cfg.OVERWRITE_ALL, cfg.CANCEL],
            )
            if response == cfg.CANCEL:
                return

    self.processes_widget = dialog.Info_widget()
    self.processes_widget.resize(700, 300)

    self.processes_widget.setWindowFlags(Qt.WindowStaysOnTopHint)
    if action == "reencode_resize":
        self.processes_widget.setWindowTitle("Re-encoding and resizing with FFmpeg")
    if action == "rotate":
        self.processes_widget.setWindowTitle("Rotating the video with FFmpeg")
    if action == "merge":
        self.processes_widget.setWindowTitle("Merging media files")

    self.processes_widget.label.setText("This operation can be long. Be patient...\nIn the meanwhile you can continue to use BORIS\n\n")
    self.processes_widget.number_of_files = len(file_names)
    self.processes_widget.show()

    if action == "merge":
        # ffmpeg -f concat -safe 0 -i join_video.txt -c copy output_demuxer.mp4
        args = ["-hide_banner", "-y", "-f", "concat", "-safe", "0", "-i", file_list, "-c", "copy", output_file_name]
        self.processes.append([QProcess(self), [self.ffmpeg_bin, args, output_file_name]])
        self.processes[-1][0].setProcessChannelMode(QProcess.MergedChannels)
        self.processes[-1][0].readyReadStandardOutput.connect(lambda: readStdOutput(len(self.processes)))
        self.processes[-1][0].readyReadStandardError.connect(lambda: readStdOutput(len(self.processes)))
        self.processes[-1][0].finished.connect(lambda: qprocess_finished(len(self.processes)))

        self.processes[-1][0].start(self.processes[-1][1][0], self.processes[-1][1][1])

    if action in ("reencode_resize", "rotate"):
        for file_name in sorted(file_names, reverse=True):
            if action == "reencode_resize":
                args = [
                    "-hide_banner",
                    "-y",
                    "-i",
                    f"{file_name}",
                    "-vf",
                    f"scale={horiz_resol}:-1",
                    "-b:v",
                    f"{video_quality * 1024 * 1024}",
                    f"{file_name}.re-encoded.{horiz_resol}px.{video_quality}Mb.avi",
                ]

            if action == "rotate":
                # check bitrate
                r = util.accurate_media_analysis(self.ffmpeg_bin, file_name)
                if "error" not in r and r["bitrate"] is not None:
                    current_bitrate = r["bitrate"]
                else:
                    current_bitrate = 10_000_000

                if rotation_idx in (1, 2):
                    args = [
                        "-hide_banner",
                        "-y",
                        "-i",
                        f"{file_name}",
                        "-vf",
                        f"transpose={rotation_idx}",
                        "-codec:a",
                        "copy",
                        "-b:v",
                        f"{current_bitrate}",
                        f"{file_name}.rotated{['', '90', '-90'][rotation_idx]}.avi",
                    ]

                if rotation_idx == 3:  # 180
                    args = [
                        "-hide_banner",
                        "-y",
                        "-i",
                        f"{file_name}",
                        "-vf",
                        "transpose=2,transpose=2",
                        "-codec:a",
                        "copy",
                        "-b:v",
                        f"{current_bitrate}",
                        f"{file_name}.rotated180.avi",
                    ]

            logging.debug("Launch process")
            logging.debug(f"{self.ffmpeg_bin} {' '.join(args)}")

            self.processes.append([QProcess(self), [self.ffmpeg_bin, args, file_name]])

            # self.processes[-1][0].setProcessChannelMode(QProcess.SeparateChannels)

            ## FFmpeg output the work in progress on stderr
            self.processes[-1][0].setProcessChannelMode(QProcess.MergedChannels)
            self.processes[-1][0].readyReadStandardOutput.connect(lambda: readStdOutput(len(self.processes)))
            # self.processes[-1][0].readyReadStandardError.connect(lambda: readStdOutput(len(self.processes)))

            self.processes[-1][0].finished.connect(lambda: qprocess_finished(len(self.processes)))

        self.processes[-1][0].start(self.processes[-1][1][0], self.processes[-1][1][1])



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

import logging
from decimal import Decimal as dec

import numpy as np
from PyQt5.QtWidgets import QInputDialog, QMessageBox

from . import config as cfg
from . import db_functions, dialog, project_functions, select_subj_behav
from . import utilities as util
from . import select_observations
from . import observation_operations


def subj_behav_modif(cursor, obsid: str, subject: str, time: dec, interval, include_modifiers: bool) -> list:
    """
    current behaviors for observation obsId at time

    Args:
        cursor (sqlite3.cursor): cursor to aggregated events db
        obsid (str): id of observation
        subject (str): name of subject
        time (Decimal): time
        include_modifiers (bool): True: include modifiers False: do not

    Returns:
        list: list of lists [subject, behavior, modifiers]
    """

    s = []
    # state behaviors
    rows = cursor.execute(
        (
            "SELECT behavior, modifiers FROM aggregated_events "
            "WHERE "
            "observation = ? "
            "AND subject = ? "
            "AND type = 'STATE' "
            "AND (? BETWEEN start AND STOP) "
        ),
        (
            obsid,
            subject,
            float(time),
        ),
    ).fetchall()

    for row in rows:
        if include_modifiers:
            s.append([subject, row[0], row[1]])
        else:
            s.append([subject, row[0]])

    # point behaviors
    rows = cursor.execute(
        (
            "SELECT behavior, modifiers FROM aggregated_events "
            "WHERE "
            "observation = ? "
            "AND subject = ? "
            "AND type = 'POINT' "
            "AND abs(start - ?) <= ? "
        ),
        (
            obsid,
            subject,
            float(time),
            float(interval / 2),
        ),
    ).fetchall()

    for row in rows:
        if include_modifiers:
            s.append([subject, row[0], row[1]])
        else:
            s.append([subject, row[0]])

    return s


def cohen_kappa(cursor, obsid1: str, obsid2: str, interval: dec, selected_subjects: list, include_modifiers: bool):
    """
    Inter-rater reliability Cohen's kappa coefficient (time-unit)
    see Sequential Analysis and Observational Methods for the Behavioral Sciences p. 77

    Args:
        cursor (sqlite3.cursor): cursor to aggregated events db
        obsid1 (str): id of observation #1
        obsid2 (str): id of observation #2
        interval (decimal.Decimal): time unit (s)
        selected_subjects (list): subjects selected for analysis
        include_modifiers (bool): True: include modifiers False: do not

    Return:
        float: K
        str: result of analysis
    """

    # check if obs have events
    for obs_id in [obsid1, obsid2]:
        if not cursor.execute("SELECT * FROM aggregated_events WHERE observation = ? ", (obs_id,)).fetchall():
            return -100, f"The observation {obs_id} has no recorded events"

    first_event = cursor.execute(
        (
            "SELECT min(start) FROM aggregated_events "
            f"WHERE observation in (?, ?) AND subject in ({','.join('?'*len(selected_subjects))}) "
        ),
        (obsid1, obsid2) + tuple(selected_subjects),
    ).fetchone()[0]

    logging.debug(f"first_event: {first_event}")

    last_event = cursor.execute(
        (
            "SELECT max(stop) FROM aggregated_events "
            f"WHERE observation in (?, ?) AND subject in ({','.join('?'*len(selected_subjects))}) "
        ),
        (obsid1, obsid2) + tuple(selected_subjects),
    ).fetchone()[0]

    logging.debug(f"last_event: {last_event}")

    nb_events1 = cursor.execute(
        ("SELECT COUNT(*) FROM aggregated_events " f"WHERE observation = ? AND subject in ({','.join('?'*len(selected_subjects))}) "),
        (obsid1,) + tuple(selected_subjects),
    ).fetchone()[0]
    nb_events2 = cursor.execute(
        ("SELECT COUNT(*) FROM aggregated_events " f"WHERE observation = ? AND subject in ({','.join('?'*len(selected_subjects))}) "),
        (obsid2,) + tuple(selected_subjects),
    ).fetchone()[0]

    total_states = []

    currentTime = dec(str(first_event))
    while currentTime <= last_event:
        for obsid in [obsid1, obsid2]:
            for subject in selected_subjects:
                s = subj_behav_modif(cursor, obsid, subject, currentTime, interval, include_modifiers)

                if s not in total_states:
                    total_states.append(s)

                logging.debug(f"{obsid} {subject} {currentTime} {s}")

        currentTime += interval

    total_states = sorted(total_states)

    logging.debug(f"total_states: {total_states} len:{len(total_states)}")

    contingency_table = np.zeros((len(total_states), len(total_states)))

    seq1 = {}
    seq2 = {}
    currentTime = dec(str(first_event))
    while currentTime <= last_event:
        seq1[currentTime] = []
        seq2[currentTime] = []
        for subject in selected_subjects:
            s1 = subj_behav_modif(cursor, obsid1, subject, currentTime, interval, include_modifiers)
            s2 = subj_behav_modif(cursor, obsid2, subject, currentTime, interval, include_modifiers)

            seq1[currentTime].append(s1)
            seq2[currentTime].append(s2)

            logging.debug(f"currentTime: {currentTime} s1:{s1} s2:{s2}")

            try:
                contingency_table[total_states.index(s1), total_states.index(s2)] += 1
            except Exception:
                return -100, "Error with contingency table"

        currentTime += interval

    logging.debug(f"seq1:\n {list(seq1.values())}")
    logging.debug(f"seq2:\n {list(seq2.values())}")

    logging.debug(f"contingency_table:\n {contingency_table}")

    template = (
        "Observation: {obsid1}\n"
        "number of events: {nb_events1}\n\n"
        "Observation: {obsid2}\n"
        "number of events: {nb_events2:.0f}\n\n"
        "K = {K:.3f}"
    )

    # out += "Observation length: <b>{:.3f} s</b><br>".format(self.observationTotalMediaLength(obsid1))
    # out += "Number of intervals: <b>{:.0f}</b><br><br>".format(self.observationTotalMediaLength(obsid1) / interval)

    # out += "Observation length: <b>{:.3f} s</b><br>".format(self.observationTotalMediaLength(obsid2))
    # out += "Number of intervals: <b>{:.0f}</b><br><br>".format(self.observationTotalMediaLength(obsid2) / interval)

    cols_sums = contingency_table.sum(axis=0)
    rows_sums = contingency_table.sum(axis=1)
    overall_total = contingency_table.sum()

    logging.debug(f"overall_total: {overall_total}")

    agreements = sum(contingency_table.diagonal())

    logging.debug(f"agreements: {agreements}")

    sum_ef = 0
    for idx in range(len(total_states)):
        sum_ef += rows_sums[idx] * cols_sums[idx] / overall_total

    logging.debug(f"sum_ef {sum_ef}")

    if not (overall_total - sum_ef):
        K = 1
    else:
        try:
            K = round((agreements - sum_ef) / (overall_total - sum_ef), 3)
        except Exception:
            K = np.nan

    out = template.format(obsid1=obsid1, obsid2=obsid2, nb_events1=nb_events1, nb_events2=nb_events2, K=K)

    logging.debug(f"K: {K}")
    return K, out


def irr_cohen_kappa(self):
    """
    calculate the Inter-Rater Reliability index - Cohen's Kappa of 2 or more observations
    https://en.wikipedia.org/wiki/Cohen%27s_kappa
    """

    # ask user observations to analyze
    _, selected_observations = select_observations.select_observations2(
        self, mode=cfg.MULTIPLE, windows_title="Select observations for IRR Cohen Kappa"
    )

    if not selected_observations:
        return
    if len(selected_observations) < 2:
        QMessageBox.information(self, cfg.programName, "Select almost 2 observations for IRR analysis")
        return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selected_observations):
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    start_coding, end_coding, _ = observation_operations.coding_time(self.pj[cfg.OBSERVATIONS], selected_observations)

    # exit with message if events do not have timestamp
    if start_coding.is_nan():
        QMessageBox.critical(
            None,
            cfg.programName,
            ("This function is not available for observations with events that do not have timestamp"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=dec("NaN"),
        end_coding=dec("NaN"),
        flagShowIncludeModifiers=True,
        flagShowExcludeBehaviorsWoEvents=False,
        n_observations=len(selected_observations),
    )
    if parameters == {}:
        return
    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        QMessageBox.warning(None, cfg.programName, "Select subject(s) and behavior(s) to analyze")
        return

    # ask for time slice
    i, ok = QInputDialog.getDouble(self, "IRR - Cohen's Kappa (time-unit)", "Time unit (in seconds):", 1.0, 0.001, 86400, 3)
    if not ok:
        return
    interval = util.float2decimal(i)

    ok, msg, db_connector = db_functions.load_aggregated_events_in_db(
        self.pj, parameters[cfg.SELECTED_SUBJECTS], selected_observations, parameters[cfg.SELECTED_BEHAVIORS]
    )

    cursor = db_connector.cursor()
    out = (
        "Index of Inter-rater Reliability - Cohen's Kappa\n\n"
        f"Interval time: {interval:.3f} s\n"
        f"Selected subjects: {', '.join(parameters[cfg.SELECTED_SUBJECTS])}\n\n"
    )

    mem_done = []
    irr_results = np.ones((len(selected_observations), len(selected_observations)))

    for obs_id1 in selected_observations:
        for obs_id2 in selected_observations:
            if obs_id1 == obs_id2:
                continue
            if set([obs_id1, obs_id2]) not in mem_done:
                K, msg = cohen_kappa(
                    cursor,
                    obs_id1,
                    obs_id2,
                    interval,
                    parameters[cfg.SELECTED_SUBJECTS],
                    parameters[cfg.INCLUDE_MODIFIERS],
                )
                irr_results[selected_observations.index(obs_id1), selected_observations.index(obs_id2)] = K
                irr_results[selected_observations.index(obs_id2), selected_observations.index(obs_id1)] = K
                out += msg + "\n=============\n"
                mem_done.append(set([obs_id1, obs_id2]))

    out2 = "\t{}\n".format("\t".join(list(selected_observations)))
    for r in range(irr_results.shape[0]):
        out2 += f"{selected_observations[r]}\t"
        out2 += "\t".join(["%8.6f" % x for x in irr_results[r, :]]) + "\n"

    self.results = dialog.Results_dialog()
    self.results.setWindowTitle("BORIS - IRR - Cohen's Kappa (time-unit) analysis results")
    self.results.ptText.setReadOnly(True)
    if len(selected_observations) == 2:
        self.results.ptText.appendPlainText(out)
    else:
        self.results.ptText.appendPlainText(out2)
    self.results.show()


def needleman_wunsch_identity(cursor, obsid1: str, obsid2: str, interval, selected_subjects: list, include_modifiers: bool):
    """
    Needleman - Wunsch identity between 2 observations

    see http://anhaidgroup.github.io/py_stringmatching/v0.4.1/NeedlemanWunsch.html#

    Args:
        cursor (sqlite3.cursor): cursor to aggregated events db
        obsid1 (str): id of observation #1
        obsid2 (str): id of observation #2
        interval
        selected_subjects (list): subjects selected for analysis
        include_modifiers (bool): True: include modifiers False: do not

    Return:
        float: identity
        str: result of analysis
    """

    def zeros(shape):
        retval = []
        for x in range(shape[0]):
            retval.append([])
            for y in range(shape[1]):
                retval[-1].append(0)
        return retval

    match_award = 1
    mismatch_penalty = -1
    gap_penalty = -1

    def match_score(alpha, beta):
        if alpha == beta:
            return match_award
        elif alpha == "-" or beta == "-":
            return gap_penalty
        else:
            return mismatch_penalty

    def finalize(align1, align2):
        align1 = align1[::-1]
        align2 = align2[::-1]

        i = 0
        symbol = []
        score = 0
        identity = 0
        for i in range(0, len(align1)):
            if align1[i] == align2[i]:
                symbol.append(align1[i])
                identity += 1
                score += match_score(align1[i], align2[i])

            elif align1[i] != align2[i] and align1[i] != "-" and align2[i] != "-":
                score += match_score(align1[i], align2[i])
                symbol.append(" ")

            # if one of them is a gap, output a space
            elif align1[i] == "-" or align2[i] == "-":
                symbol.append(" ")
                score += gap_penalty

        identity = float(identity) / len(align1) * 100

        return {"identity": identity, "score": score, "align1": align1, "align2": align2, "symbol": symbol}

    def needle(seq1, seq2):
        m, n = len(seq1), len(seq2)

        score = zeros((m + 1, n + 1))

        for i in range(0, m + 1):
            score[i][0] = gap_penalty * i
        for j in range(0, n + 1):
            score[0][j] = gap_penalty * j
        for i in range(1, m + 1):
            for j in range(1, n + 1):
                match = score[i - 1][j - 1] + match_score(seq1[i - 1], seq2[j - 1])
                delete = score[i - 1][j] + gap_penalty
                insert = score[i][j - 1] + gap_penalty
                score[i][j] = max(match, delete, insert)

        align1, align2 = [], []
        i, j = m, n
        while i > 0 and j > 0:
            score_current = score[i][j]
            score_diagonal = score[i - 1][j - 1]
            score_up = score[i][j - 1]
            score_left = score[i - 1][j]

            if score_current == score_diagonal + match_score(seq1[i - 1], seq2[j - 1]):
                align1.append(seq1[i - 1])
                align2.append(seq2[j - 1])
                i -= 1
                j -= 1
            elif score_current == score_left + gap_penalty:
                align1.append(seq1[i - 1])
                align2.append("-")
                i -= 1
            elif score_current == score_up + gap_penalty:
                align1.append("-")
                align2.append(seq2[j - 1])
                j -= 1

        # Finish tracing up to the top left cell
        while i > 0:
            align1.append(seq1[i - 1])
            align2.append("-")
            i -= 1
        while j > 0:
            align1.append("-")
            align2.append(seq2[j - 1])
            j -= 1

        return finalize(align1, align2)

    first_event = cursor.execute(
        (
            "SELECT min(start) FROM aggregated_events "
            f"WHERE observation in (?, ?) AND subject in ({','.join('?'*len(selected_subjects))}) "
        ),
        (obsid1, obsid2) + tuple(selected_subjects),
    ).fetchone()[0]

    if first_event is None:
        logging.debug(f"An observation has no recorded events: {obsid1} or {obsid2}")

        return -100, f"An observation has no recorded events: {obsid1} {obsid2}"

    logging.debug(f"first_event: {first_event}")

    last_event = cursor.execute(
        (
            "SELECT max(stop) FROM aggregated_events "
            f"WHERE observation in (?, ?) AND subject in ({','.join('?'*len(selected_subjects))}) "
        ),
        (obsid1, obsid2) + tuple(selected_subjects),
    ).fetchone()[0]

    logging.debug(f"last_event: {last_event}")

    nb_events1 = cursor.execute(
        ("SELECT COUNT(*) FROM aggregated_events " f"WHERE observation = ? AND subject in ({','.join('?'*len(selected_subjects))}) "),
        (obsid1,) + tuple(selected_subjects),
    ).fetchone()[0]

    nb_events2 = cursor.execute(
        ("SELECT COUNT(*) FROM aggregated_events " f"WHERE observation = ? AND subject in ({','.join('?'*len(selected_subjects))}) "),
        (obsid2,) + tuple(selected_subjects),
    ).fetchone()[0]

    seq1: dict = {}
    seq2: dict = {}

    currentTime = dec(str(first_event))
    while currentTime <= last_event:
        seq1[currentTime], seq2[currentTime] = [], []

        for subject in selected_subjects:
            s1 = subj_behav_modif(cursor, obsid1, subject, currentTime, interval, include_modifiers)
            s2 = subj_behav_modif(cursor, obsid2, subject, currentTime, interval, include_modifiers)

            seq1[currentTime].append(s1)
            seq2[currentTime].append(s2)

            logging.debug(f"currentTime: {currentTime} s1:{s1} s2:{s2}")

        currentTime += interval

    logging.debug(f"seq1:\n {list(seq1.values())}")
    logging.debug(f"seq2:\n {list(seq2.values())}")

    r = needle(list(seq1.values()), list(seq2.values()))

    out = (
        f"Observation: {obsid1}\n"
        f"number of events: {nb_events1}\n\n"
        f"Observation: {obsid2}\n"
        f"number of events: {nb_events2:.0f}\n\n"
        f"identity = {r['identity']:.3f} %"
    )

    logging.debug(f"identity: {r['identity']}")

    return r["identity"], out


def needleman_wunch(self):
    """
    calculate the Needleman-Wunsch similarity for 2 or more observations
    """

    # ask user observations to analyze
    _, selected_observations = select_observations.select_observations2(
        self, mode=cfg.MULTIPLE, windows_title="Select observations for Needleman-Wunch identity"
    )

    if not selected_observations:
        return
    if len(selected_observations) < 2:
        QMessageBox.information(self, cfg.programName, "You have to select at least 2 observations for Needleman-Wunsch similarity")
        return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selected_observations):
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    start_coding, end_coding, _ = observation_operations.coding_time(self.pj[cfg.OBSERVATIONS], selected_observations)

    # exit with message if events do not have timestamp
    if start_coding.is_nan():
        QMessageBox.critical(
            None,
            cfg.programName,
            ("This function is not available for observations with events that do not have timestamp"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=dec("NaN"),
        end_coding=dec("NaN"),
        flagShowIncludeModifiers=True,
        flagShowExcludeBehaviorsWoEvents=False,
        n_observations=len(selected_observations),
    )

    if parameters == {}:
        return

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        QMessageBox.warning(None, cfg.programName, "Select subject(s) and behavior(s) to analyze")
        return

    # ask for time slice

    i, ok = QInputDialog.getDouble(self, "Needleman-Wunsch similarity", "Time unit (in seconds):", 1.0, 0.001, 86400, 3)
    if not ok:
        return
    interval = util.float2decimal(i)

    ok, msg, db_connector = db_functions.load_aggregated_events_in_db(
        self.pj, parameters[cfg.SELECTED_SUBJECTS], selected_observations, parameters[cfg.SELECTED_BEHAVIORS]
    )

    cursor = db_connector.cursor()
    out = (
        "Needleman-Wunsch similarity\n\n"
        f"Time unit: {interval:.3f} s\n"
        f"Selected subjects: {', '.join(parameters[cfg.SELECTED_SUBJECTS])}\n\n"
    )
    mem_done = []
    nws_results = np.ones((len(selected_observations), len(selected_observations)))

    for obs_id1 in selected_observations:
        for obs_id2 in selected_observations:
            if obs_id1 == obs_id2:
                continue
            if set([obs_id1, obs_id2]) not in mem_done:
                similarity, msg = needleman_wunsch_identity(
                    cursor,
                    obs_id1,
                    obs_id2,
                    interval,
                    parameters[cfg.SELECTED_SUBJECTS],
                    parameters[cfg.INCLUDE_MODIFIERS],
                )
                nws_results[selected_observations.index(obs_id1), selected_observations.index(obs_id2)] = similarity
                nws_results[selected_observations.index(obs_id2), selected_observations.index(obs_id1)] = similarity
                out += msg + "\n=============\n"
                mem_done.append(set([obs_id1, obs_id2]))

    out2 = "\t{}\n".format("\t".join(list(selected_observations)))
    for r in range(nws_results.shape[0]):
        out2 += f"{selected_observations[r]}\t"
        out2 += "\t".join([f"{x:8.6f}" for x in nws_results[r, :]]) + "\n"

    self.results = dialog.Results_dialog()
    self.results.setWindowTitle(f"{cfg.programName} - Needleman-Wunsch similarity")
    self.results.ptText.setReadOnly(True)
    if len(selected_observations) == 2:
        self.results.ptText.appendPlainText(out)
    else:
        self.results.ptText.appendPlainText(out2)
    self.results.show()



"""
BORIS CLI

Behavioral Observation Research Interactive Software Command Line Interface

Copyright 2012-2024 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

import argparse
import sys
import re
import pathlib
import utilities
import project_functions
from config import *
import db_functions
import export_observation
import irr
import plot_events
import version

__version__ = version.__version__
__version_date__ = version.__version_date__


def cleanhtml(raw_html):
    raw_html = raw_html.replace("<br>", "\n")
    cleanr = re.compile("<.*?>")
    cleantext = re.sub(cleanr, "", raw_html)
    return cleantext


def all_observations(pj):
    return [idx for idx in sorted(pj[OBSERVATIONS])]


commands_list = ["check_state_events", "export_events", "irr", "subtitles", "check_project_integrity", "plot_events"]
commands_usage = {
    "check_state_events": (
        "usage:\nboris_cli -p PROJECT_FILE -o OBSERVATION_ID --command check_state_events\n"
        "where\n"
        "PROJECT_FILE is the path of the BORIS project\n"
        "OBSERVATION_ID is the id of observation(s) (if ommitted all observations are checked)"
    ),
    "export_events": (
        "usage:\nboris_cli -p PROJECT_FILE -o OBSERVATION_ID --command export_events [OUTPUT_FORMAT]\n"
        "where:\n"
        "PROJECT_FILE is the path of the BORIS project\n"
        "OBSERVATION_ID is the id of observation(s) (if ommitted all observations are exported)\n"
        "OUTPUT_FORMAT can be tsv (default), csv, xls, xlsx, ods, html"
    ),
    "irr": (
        'usage:\nboris_cli -p PROJECT_FILE -o "OBSERVATION_ID1" "OBSERVATION_ID2" --command irr [INTERVAL] [INCLUDE_MODIFIERS]\n'
        "where:\n"
        "PROJECT_FILE is the path of the BORIS project\n"
        "INTERVAL in seconds (default is 1)\n"
        "INCLUDE_MODIFIERS must be true or false (default is true)"
    ),
    "subtitles": (
        'usage:\nboris_cli -p PROJECT_FILE -o "OBSERVATION_ID" --command subtitles [OUTPUT_DIRECTORY]\n'
        "where:\n"
        "OUTPUT_DIRECTORY is the directory where subtitles files will be saved"
    ),
    "check_project_integrity": "usage:\nboris_cli -p PROJECT_FILE --command check_project_integrity",
    "plot_events": (
        "usage:\nboris_cli - p PROJECT_FILE -o OBSERVATION_ID --command plot_events "
        "[OUTPUT_DIRECTORY] [INCLUDE_MODIFIERS] [EXCLUDE_BEHAVIORS] [PLOT_FORMAT]\n"
        "where\n"
        "OUTPUT_DIRECTORY is the directory where the plots will be saved\n"
        "INCLUDE_MODIFIERS must be true or false (default is true)\n"
        "EXCLUDE_BEHAVIORS: True: behaviors without events are not plotted (default is true)\n"
        "PLOT_FORMAT can be png, svg, pdf, ps"
    ),
}

parser = argparse.ArgumentParser(description="BORIS CLI")
parser.add_argument("-v", "--version", action="store_true", dest="version", help="BORIS version")
parser.add_argument("-p", "--project", action="store", dest="project_file", help="Project file path")
parser.add_argument("-o", "--observation", nargs="*", action="store", default=[], dest="observation_id", help="Observation id")
parser.add_argument("-i", "--info", action="store_true", dest="project_info", help="Project information")
parser.add_argument("-c", "--command", nargs="*", action="store", dest="command", help="Command to execute")

args = parser.parse_args()

pj, observations_id_list = {}, {}

if args.version:
    print("version {}".format(__version__))
    sys.exit()

if args.command:
    if args.command[0].upper() == "LIST":
        for command in commands_list:
            print(command)
            print("=" * len(command))
            if command in commands_usage:
                print(commands_usage[command])
            print()
        print()
        sys.exit()

if args.project_file:
    if not args.command:
        print("Project path: {}".format(args.project_file))

    project_path, project_changed, pj, msg = project_functions.open_project_json(args.project_file)
    if "error" in pj:
        print(pj["error"])
        sys.exit()
    if msg:
        print(msg)

if args.observation_id:
    observations_id_list = args.observation_id
    """
    if not args.command:
        print("\nObservations:")
        for observation_id in observations_id_list:
            if observation_id in pj[OBSERVATIONS]:
                print("Id: {}".format(observation_id))
            else:
                print("{}: NOT FOUND in project".format(observation_id))
        print()
    """

if args.project_info:
    if not args.command:
        if pj:
            print("Project name: {}".format(pj[PROJECT_NAME]))
            print("Project date: {}".format(pj[PROJECT_DATE]))
            print("Project description: {}".format(pj[PROJECT_DESCRIPTION]))
            print()

            if not observations_id_list:
                print("Ethogram\n========")
                print("Number of behaviors in ethogram: {}".format(len(pj[ETHOGRAM])))
                for idx in utilities.sorted_keys(pj[ETHOGRAM]):
                    print(
                        "Code: {}\tDescription: {}\tType: {}".format(
                            pj[ETHOGRAM][idx][BEHAVIOR_CODE], pj[ETHOGRAM][idx]["description"], pj[ETHOGRAM][idx][TYPE]
                        )
                    )
                """print("Behaviors: {}".format(",".join([pj[ETHOGRAM][k]["code"] for k in utilities.sorted_keys(pj[ETHOGRAM])])))"""
                print()

                print("Subjects\n========")
                print("Number of subjects: {}".format(len(pj[SUBJECTS])))
                for idx in utilities.sorted_keys(pj[SUBJECTS]):
                    print("Name: {}\tDescription: {}".format(pj[SUBJECTS][idx]["name"], pj[SUBJECTS][idx]["description"]))
                print()

                print("Observations\n============")
                print("Number of observations: {}".format(len(pj[OBSERVATIONS])))
                print("List of observations:")
                for observation_id in sorted(pj[OBSERVATIONS].keys()):
                    print("Id: {}\tDate: {}".format(observation_id, pj[OBSERVATIONS][observation_id]["date"]))

            else:
                for observation_id in observations_id_list:
                    print("Observation id: {}".format(observation_id))
                    if pj[OBSERVATIONS][observation_id][EVENTS]:
                        for event in pj[OBSERVATIONS][observation_id][EVENTS]:
                            print("\t".join([str(x) for x in event]))
                    else:
                        print("No events recorded")
                    print()
        else:
            print("No project")
        sys.exit()

if args.command:
    print("Command: {}\n".format(" ".join(args.command)))

    if not pj:
        print("No project")
        sys.exit()

    if "check_state_events" in args.command:
        if not observations_id_list:
            print("No observation selected. Command applied on all observations found in project\n")
            observations_id_list = all_observations(pj)

        for observation_id in observations_id_list:
            ret, msg = project_functions.check_state_events_obs(observation_id, pj[ETHOGRAM], pj[OBSERVATIONS][observation_id], HHMMSS)
            print("{}: {}".format(observation_id, cleanhtml(msg)))
        sys.exit()

    if "export_events" in args.command:
        if not observations_id_list:
            print("No observation selected. Command applied on all observations found in project\n")
            observations_id_list = [idx for idx in pj[OBSERVATIONS]]

        behaviors = [pj[ETHOGRAM][k]["code"] for k in utilities.sorted_keys(pj[ETHOGRAM])]
        subjects = [pj[SUBJECTS][k]["name"] for k in utilities.sorted_keys(pj[SUBJECTS])] + [NO_FOCAL_SUBJECT]

        output_format = "tsv"
        if len(args.command) > 1:
            output_format = args.command[1]

        for observation_id in observations_id_list:
            ok, msg = export_observation.export_events(
                {"selected subjects": subjects, "selected behaviors": behaviors},
                observation_id,
                pj[OBSERVATIONS][observation_id],
                pj[ETHOGRAM],
                utilities.safeFileName(observation_id + "." + output_format),
                output_format,
            )
            if not ok:
                print(msg)

        sys.exit()

    if "irr" in args.command:
        if len(observations_id_list) != 2:
            print("select 2 observations")
            sys.exit()

        behaviors = [pj[ETHOGRAM][k]["code"] for k in utilities.sorted_keys(pj[ETHOGRAM])]
        subjects = [pj[SUBJECTS][k]["name"] for k in utilities.sorted_keys(pj[SUBJECTS])] + [NO_FOCAL_SUBJECT]

        ok, msg, db_connector = db_functions.load_aggregated_events_in_db(pj, subjects, observations_id_list, behaviors)

        if not ok:
            print(cleanhtml(msg))
            sys.exit()

        cursor = db_connector.cursor()

        interval = 1
        if len(args.command) > 1:
            interval = utilities.float2decimal(args.command[1])

        include_modifiers = True
        if len(args.command) > 2:
            include_modifiers = "TRUE" in args.command[2].upper()

        K, out = irr.cohen_kappa(cursor, observations_id_list[0], observations_id_list[1], interval, subjects, include_modifiers)

        print(("Cohen's Kappa - Index of Inter-Rater Reliability\n\n" "Interval time: {interval:.3f} s\n").format(interval=interval))

        print(out)
        sys.exit()

    if "subtitles" in args.command:
        if not observations_id_list:
            print("No observation selected. Command applied on all observations found in project\n")
            observations_id_list = all_observations(pj)

        behaviors = [pj[ETHOGRAM][k]["code"] for k in utilities.sorted_keys(pj[ETHOGRAM])]
        subjects = [pj[SUBJECTS][k]["name"] for k in utilities.sorted_keys(pj[SUBJECTS])] + [NO_FOCAL_SUBJECT]

        export_dir = "."
        if len(args.command) > 1:
            export_dir = args.command[1]
            if not pathlib.Path(export_dir).is_dir():
                print("{} is not a valid directory".format(export_dir))
                sys.exit()

        ok, msg = project_functions.create_subtitles(
            pj,
            observations_id_list,
            {"selected subjects": subjects, "selected behaviors": behaviors, "include modifiers": True},
            export_dir,
        )
        if not ok:
            print(cleanhtml(msg))
        sys.exit()

    if "check_project_integrity" in args.command[0]:
        msg = project_functions.check_project_integrity(pj, HHMMSS, args.project_file)
        if msg:
            print(cleanhtml(msg))
        else:
            print("No issuses found in project")
        sys.exit()

    if "plot_events" in args.command[0]:
        if not observations_id_list:
            print("No observation selected. Command applied on all observations found in project\n")
            observations_id_list = all_observations(pj)

        behaviors = [pj[ETHOGRAM][k]["code"] for k in utilities.sorted_keys(pj[ETHOGRAM])]
        subjects = [pj[SUBJECTS][k]["name"] for k in utilities.sorted_keys(pj[SUBJECTS])] + [NO_FOCAL_SUBJECT]

        export_dir = "."
        if len(args.command) > 1:
            export_dir = args.command[1]
            if not pathlib.Path(export_dir).is_dir():
                print("{} is not a valid directory".format(export_dir))
                sys.exit()

        include_modifiers = True
        if len(args.command) > 2:
            include_modifiers = "TRUE" in args.command[2].upper()

        exclude_behaviors = True
        if len(args.command) > 3:
            exclude_behaviors = False if "FALSE" in args.command[3].upper() else True

        plot_format = "png"
        if len(args.command) > 4:
            plot_format = args.command[4].lower()

        plot_events.create_events_plot(
            pj,
            observations_id_list,
            {
                "selected subjects": subjects,
                "selected behaviors": behaviors,
                "include modifiers": include_modifiers,
                "exclude behaviors": exclude_behaviors,
                "time": TIME_FULL_OBS,
                "start time": 0,
                "end time": 0,
            },
            plot_colors=BEHAVIORS_PLOT_COLORS,
            plot_directory=export_dir,
            file_format=plot_format,
        )
        sys.exit()

    print("Command {} not found!".format(args.command[0]))

print()



# -*- coding: utf-8 -*-

"""Base palette mixin."""

# Standard library imports
from collections import OrderedDict


class Palette(object):

    ID = None

    # Color
    COLOR_BACKGROUND_1 = ''
    COLOR_BACKGROUND_2 = ''
    COLOR_BACKGROUND_3 = ''
    COLOR_BACKGROUND_4 = ''
    COLOR_BACKGROUND_5 = ''
    COLOR_BACKGROUND_6 = ''

    COLOR_TEXT_1 = ''
    COLOR_TEXT_2 = ''
    COLOR_TEXT_3 = ''
    COLOR_TEXT_4 = ''

    COLOR_ACCENT_1 = ''
    COLOR_ACCENT_2 = ''
    COLOR_ACCENT_3 = ''
    COLOR_ACCENT_4 = ''
    COLOR_ACCENT_5 = ''

    OPACITY_TOOLTIP = 0

    # Size
    SIZE_BORDER_RADIUS = '4px'

    # Borders
    BORDER_1 = '1px solid $COLOR_BACKGROUND_1'
    BORDER_2 = '1px solid $COLOR_BACKGROUND_4'
    BORDER_3 = '1px solid $COLOR_BACKGROUND_6'

    BORDER_SELECTION_3 = '1px solid $COLOR_ACCENT_3'
    BORDER_SELECTION_2 = '1px solid $COLOR_ACCENT_2'
    BORDER_SELECTION_1 = '1px solid $COLOR_ACCENT_1'

    # Example of additional widget specific variables
    W_STATUS_BAR_BACKGROUND_COLOR = COLOR_ACCENT_1

    # Paths
    PATH_RESOURCES = "':/qss_icons'"

    @classmethod
    def to_dict(cls, colors_only=False):
        """Convert variables to dictionary."""
        order = [
            'ID',
            'COLOR_BACKGROUND_6',
            'COLOR_BACKGROUND_5',
            'COLOR_BACKGROUND_4',
            'COLOR_BACKGROUND_2',
            'COLOR_BACKGROUND_3',
            'COLOR_BACKGROUND_1',
            'COLOR_TEXT_1',
            'COLOR_TEXT_2',
            'COLOR_TEXT_3',
            'COLOR_TEXT_4',
            'COLOR_ACCENT_1',
            'COLOR_ACCENT_2',
            'COLOR_ACCENT_3',
            'COLOR_ACCENT_4',
            'COLOR_DISABLED',
            'OPACITY_TOOLTIP',
            'SIZE_BORDER_RADIUS',
            'BORDER_1',
            'BORDER_2',
            'BORDER_3',
            'BORDER_SELECTION_3',
            'BORDER_SELECTION_2',
            'BORDER_SELECTION_1',
            'W_STATUS_BAR_BACKGROUND_COLOR',
            'PATH_RESOURCES',
        ]
        dic = OrderedDict()
        for var in order:
            value = getattr(cls, var)

            if var == 'ID':
                value = "'{}'".format(value)

            if colors_only:
                if not var.startswith('COLOR'):
                    value = None

            if value:
                dic[var] = value

        return dic

    @classmethod
    def color_palette(cls):
        """Return the ordered colored palette dictionary."""
        return cls.to_dict(colors_only=True)
